/* Copyright (c) 2009 NICTA 
 * All rights reserved. 
 * 
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.1 (the "License"); you may not use this file except in 
 * compliance with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/ 
 * 
 * Software distributed under the License is distributed on an "AS IS" 
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the 
 * License for the specific language governing rights and limitations 
 * under the License. 
 * 
 * Authors: Xinhua Zhang (xinhua.zhang81@gmail.com)
 *
 * Created: (19/05/2009) 
 *
 * Last Updated: (20/05/2009)   
 */

#ifndef _Nesterov03_HPP_
#define _Nesterov03_HPP_

#include <string>
#include "common.hpp"
#include "solver.hpp"
#include "bmrminnersolver.hpp"
#include "model.hpp"
#include "loss.hpp"
#include "sml.hpp"
#include "timer.hpp"
#include "hingeloss.hpp"

/**   Class for CNesterov03 solver.
 *    This type of solver iteratively builds up a convex lower-bound of the 
 *      objective function, and performs minimization on the lower-bound.
 */
class CNesterov03 : public CSolver
{
   public:
      /** Constructor
       *  @param model [r/w] model object
       *  @param loss [read] loss object
       */
      CNesterov03(CModel* model, CHingeLoss* loss);
      
      // Destructor
      virtual ~CNesterov03();
      
      // Methods
      virtual void Train();

   protected:

      TheMatrix curGrad, alphak, wk, beta, offset, best_alpha, best_w;
      double L;     // L only needs to take care of X^T X, don't worry about scaling
      unsigned int num_example, num_feature;
      double maxDualObj, minPrimalObj;
      bool useBias;
      double varScaleFactor03, vscaleRaw;

      ublas::vector<double> dsq1, dsq2;
      ublas::vector<double> lower_bound, upper_bound, sigma;
      ublas::vector<double> solution, offset_vec;
      double z;

      /** Verbosity level
       */
      int verbosity;
      
      /** Maximum number of CNesterov03 iteration
       */
      unsigned int maxNumOfIter;
            
      /** Regularization constant
       */
      double lambda;
      
      /** Prefix for intermediate/checkpoint model files
       */
      std::string checkpointPrefix;
      
      /** The number of iterations before saving a checkpoint model
       */
      unsigned int checkpointInterval;
      
      /** Selected type of checkpoint.
       */
      unsigned int checkpointMode;   
      
      /** Types of checkpoint mode.
       *  KEEP_ALL -- Keep a checkpoint model after every $checkpointInterval$ 
       *              iterations
       *  KEEP_LATEST -- Keep only the latest model
       */
      enum CHECKPOINT_MODE {KEEP_ALL, KEEP_LATEST};
      
      
      /** Validate all provided program parameters are good to use.
       *  E.g. lambda must be > 0.
       */
      virtual void ConfirmProgramParameters();

      
      /** Display iteration information
       */
      virtual void DisplayIterationInfo(unsigned int iter, double curTime, double dualObj, double primalObj);
      

      /** Save model at every #checkpointInterval iterations
       */
      virtual void SaveCheckpointModel(unsigned int iter);
         
      
      /** Termination criteria check
       *
       */
      virtual int CheckTermination(unsigned int iter);


      /** Display information after training is done
       */
      virtual void  DisplayAfterTrainingInfo(unsigned int iter, CTimer& lossAndGradientTime,
                                             CTimer& innerSolverTime, CTimer& totalTime);

      void checkVarValid(TheMatrix &var, ublas::vector<double> &lower_bound, ublas::vector<double> &upper_bound, 
                            ublas::vector<double> &sigma, double z, char * varName);

      void map_alpha_to_w(TheMatrix &alpha, TheMatrix &w)
      {
          _loss->map_alpha_to_w (alpha, w, 1.0 / lambda / num_example / varScaleFactor03);
      }

      void init_alpha(TheMatrix &alpha);
      void map_w_to_alpha(TheMatrix &wk, TheMatrix &result, double mu);
      void map_alpha_to_alpha(TheMatrix &alpha, TheMatrix &result);
};

#endif
